//---------------------------------------------------------------------------
#ifndef LogimpH
#define LogimpH
//---------------------------------------------------------------------------
/*
XPERT LOG

This file contains the actual implementation details of the Logger class.
The implementation is not exposed to the user of the DLL so that anything can
be changed in the implementation without requiring a re-compilation of programs
which use the DLL. The data structures which describe the LOG file or the LOG class
need never be distributed in order to use the LOG DLL.

Version 1.0, Jonathan D. Weisberg     (C)1999 Sutron Corp

V1.0, 2/22/2001, Supports 'Flags' field to support option to disable viewing of quality data

Implements a wrap-around time stamped data store called LOG, and CURSOR class for navigating the log.

FEATURES:
o All data is compressed using LZO compression.
o Stores binary or ascii data of variable length up to DEFAULTMAXDATASIZE.
o Access data by line number, or search by time stamp.
o Space is pre-allocated and never exceeded.
o Stores all data in 1 or more sub-blocks (DEFAULTBLOCKSIZE) to make it easy and quick to find data.
o Reads the size of data blocks and sub blocks from the LOG file itself to support future tweaks to this sizes.
o Access to the log thru the class is thread-safe.
o Compressed data is validated with a CRC16
o Error recovery is possible since one bad block does not effect the ability to read other blocks.

STRUCTURE (using current default lengths)

A LOG file consists of a 128 byte header, followed by a series of 128 byte blocks. Each block has a 1 byte
identifier which indicates the type of data stored in the block. Data of greater than 128 bytes is stored in a data
block followed by one or more continuation blocks. The first block has extra information including the
number of lines and the first timestamp. This allows for navigation without decompressing the data.
The data stored in a block has a 2 byte length before each line followed by a timestamp followed by the actual data.
The final length is FFFF to indicate no more data.

[ HEADER - 128 bytes     ]
[ DATA BLOCK - 128 bytes ]
[ CONT BLOCK - 128 bytes ]
[ CONT BLOCK - 128 bytes ]
[ CONT BLOCK - 128 bytes ]
[ DATA BLOCK - 128 bytes ]
[ CONT BLOCK - 128 bytes ]
[ END BLOCK  - 128 bytes ]


HISTORY
07/12/1999 JDW Initial Coding.

*/

/*
Initial Benchmark Data:

DEMO0101 has 4,704 lines of data, or 49 Days worth
It contains 7 sensors collected every 15 minutes or 96 lines per day

FILE NAME     SIZE      BYTES PER LINE   COMP   DESCRIPTION
------------  --------  --------------   -----  --------------------------------------------------------------
DEMO0101.LOG   66,108     14             7.7:1  8200 binary format (16 bit fixed point storage)
EXPER.LOG     108,000     23             4.7:1  Experimental NG LOG (sensor names are not stored in log)
DEMO0101.DAT  131,964     28             3.8:1  PC Base 2 binary format (IEEE single precision)
XPERT.LOG     170,000     36             3.0:1  NextGen LOG sized to just hold the DEMO0101 Data
DEMO0101.PRN  254,973     54             2.0:1  LOGPRN converted ASCII format
DEMO0101.CSV  509,413    108             1.0:1  ASCII format with "YYMMDD,HHMM,Name,Data,Name,Data..." format


DEFAULTMAXDATASIZE Approx Lines Stored (200,000 byte file w/DEFAULTBLOCKSIZE=128)     BYTES PER LINE
-----------------  -------------------------------------------------------------      --------------
1024               4167  (43 days)                                                    48
2048               5522  (57 days)                                                    36
4096               6389  (66 days)                                                    31
8192               6899  (71 days)                                                    29
16384              7104  (74 days)                                                    28
(*) Note large block sizes have many disadvantages and the moderate increase in storage would not seem worth it.

DEFAULTBLOCKSIZE   Approx Lines Stored (200,000 byte file w/DEFAULTMAXDATASIZE=2048)  BYTES PER LINE
-----------------  -----------------------------------------------------------------  --------------
32                 5881  (61 days)                                                     34
64                 5831  (60 days)                                                     34
96                 5711  (59 days)                                                     35
128                5522  (57 days)                                                     36
192                5297  (55 days)                                                     38
256                4669  (48 days)                                                     43
(*) Note the DEFAULTBLOCKSIZE doesn't start to have a huge effect until it climbs over 256 bytes


DEFAULTBLOCKSIZE   Approx Lines Stored (200,000 byte file w/DEFAULTMAXDATASIZE=16384)  BYTES PER LINE
-----------------  -----------------------------------------------------------------  --------------
64                 7104  (74 days)                                                     28
256                6976  (72 days)                                                     29
(*) Note with the large MAXDATASIZE how little of an effect the DEFAULTBLOCKSIZE has.
*/


#ifdef __BORLANDC__
typedef char TCHAR;
#else
#include <windows.h>
#endif

typedef short int INT16;
typedef unsigned short int UINT16;
typedef __int64 INT64;
typedef unsigned __int64 UINT64;
typedef unsigned char UINT8;
typedef int INT32;
typedef unsigned int UINT32;

typedef UINT16 LOGLINELEN;

const INT64 MAXTIME=140737488355327;  // 2^47-1
const INT64 MINTIME=-140737488355328; // -2^47
const int MAXBLOCKLINES=128;
const int DEFAULTBLOCKSIZE=128;
const int DEFAULTMAXDATASIZE=2048;

typedef UINT8* DATAARRAY;

// Interesting note - enums can be byte wide in Borland C but are always 32-bit in Microsoft ... so no enums for this:
typedef UINT8 BLOCKDESCRIPTOR;
const int NULLBLOCK=0;
const int DATABLOCK=1;
const int CONTBLOCK=2;
const int LASTBLOCKFLAG=128;
const int ENDOFFILE=255;
const int BLOCKMASK = ~LASTBLOCKFLAG;

// Borland and Microsoft handle structure alignment pragmas differently
#ifdef __BORLANDC__
#pragma option -a1
#else
#pragma pack(1)
#endif

class LOGTIMESTAMP
{
public:
   void Set(INT64 Time)     { memcpy(&Low, &Time, 6);                             };
   INT64 Get()              { return (INT64) High * (INT64) 65536 + (INT64) Low;  };
private:
   UINT16 Low;
   INT32 High;
};

typedef struct
{
   LOGLINELEN Len;
   LOGTIMESTAMP Time;
   UINT8 Data[DEFAULTMAXDATASIZE];
} * PLOGLINE;

struct LOGHEADER
{
   char ID[8];                    // XPERTLOG
   char Seperator[2];             // Seperator so when log file is typed it says something
   char Memo[64];                 // Note or memo
   char EOFCHAR;                  // ^Z
   UINT8 MajorVersion;            // 1
   UINT8 MinorVersion;            // 1
   UINT8 Flags;                   // varies, unused bits are 0
   UINT16 BlockSize;              // 128
   UINT16 MaxDataSize;            // 2048
   UINT16 MaxDataWrite;           // 1910
   UINT16 wMaxBlocks;             // Number of blocks stored
   UINT16 ExtraState;             // 0 = Undefined, 1=MaxBlocks is defined, 2=All extra state information is defined
   int MaxBlocks;                 // Extra State Information to speedup loads
   int FirstBlockIndex;           // Extra State Information to speedup loads
   int LastBlockIndex;            // Extra State Information to speedup loads
   int BlockCount;                // Extra State Information to speedup loads
   int LineCount;                 // Extra State Information to speedup loads
   char Reserved[20];             // first 128 bytes are header info
};

typedef PLOGLINE LINEARRAY[MAXBLOCKLINES];

class LOGBLOCK
{
public:
   int GetLine()                    { return IsData()? LineCount : 0;                                   };
   void SetLine(int Num)            { LineCount = (UINT8) Num;                                          };
   INT64 GetTime()                  { return Time.Get();                                                };
   void SetTime(INT64 time)         { Time.Set(time);                                                   };
   bool IsData()                    { return (Descriptor & BLOCKMASK) == DATABLOCK;                     };
   bool IsEnd()                     { return Descriptor == ENDOFFILE;                                   };
   bool IsContinuation()            { return (Descriptor & BLOCKMASK) == CONTBLOCK;                     };
   bool IsLast()                    { return !IsEnd() && (Descriptor & LASTBLOCKFLAG);                  };
   UINT8* Data()                    { return IsContinuation() ? &ContData[0] : &FirstData[0];           };
private:
   friend class LOGIMP;
   BLOCKDESCRIPTOR Descriptor;
   union
   {
      UINT8 ContData[DEFAULTBLOCKSIZE-sizeof(BLOCKDESCRIPTOR)];
      struct
      {
         UINT8 LineCount;
         LOGTIMESTAMP Time;
         UINT8 FirstData[DEFAULTBLOCKSIZE-sizeof(BLOCKDESCRIPTOR)-sizeof(UINT8)-sizeof(LOGTIMESTAMP)];
      };
   };
};

#ifdef __BORLANDC__
#pragma option -a4
#else
#pragma pack()
#endif

#include "LogFlags.h"

const int DEFAULTMAXDATAWRITE=DEFAULTMAXDATASIZE-DEFAULTBLOCKSIZE-sizeof(UINT16)-sizeof(UINT16)-sizeof(LOGTIMESTAMP);  // Reserve space for a length, a timestamp, and a terminating FFFF

typedef LOGBLOCK* PLOGBLOCK;

class LOGIMP
{
public:
   LOGIMP();
   ~LOGIMP();
   bool Create(LPCTSTR Name, LPCTSTR Memo, int Size, UINT32 LogFlags, int SubBlockSize=DEFAULTBLOCKSIZE, int DesiredDataSize=DEFAULTMAXDATAWRITE, int MaxDataSize=DEFAULTMAXDATASIZE);
   bool Open(LPCTSTR Name);
   bool LOGIMP::Resize(LPCTSTR Name, LPCTSTR Memo, int NewSize, UINT32 NewFlags);
   void Flush(bool Force=true, bool WriteToDisk=false);
   void SaveState();
   void ClearState();
   void Close();
   bool Append(INT64 Time, void* Data, int Len);
   bool AppendNote(INT64 Time, char* Note);
   bool AppendNote(INT64 Time, LPCTSTR Note);
   bool AppendSensor(const INT64 Time, LPCTSTR Name, LPCTSTR Units, UINT8 Quality, const double Num);
   bool AppendSensor(const INT64 Time, LPCTSTR Name, LPCTSTR Units, UINT8 Quality, LPCTSTR Num);
   bool AppendRecord(INT64 Time, LPCTSTR Id, LPCTSTR Record);
   bool Changed(UINT32& Counter);
   bool ReadExtra(PVOID Extra, int Size);
   bool WriteExtra(PVOID Extra, int Size);
   UINT32 GetFlags();
   void EraseData();
   UINT32 GetExtraSize();
   UINT32 GetSize();
   UINT32 GetLineCount();
   void GetMemo(TCHAR* Memo);
   void SetMemo(LPCTSTR Memo);
   static bool IsValidLogFile(LPCTSTR Name);                  // Returns true if name is a valid .log file
   bool BeingAccessed();
   bool BeingClosed();
   void Closing();
private:
   friend class LOGCURSORIMP;
   friend class LOGLOCK;
   PLOGBLOCK GetBlock(int Num);
   void SetBlock(int Num);
   void Setup(LOGHEADER& Header);
   bool ReadBlock(int Index, int& StartIndex, int& StopIndex, DATAARRAY Data, int& DataLen, INT64& BlockTime);
   void ReadLines(DATAARRAY Data, LINEARRAY& Lines, int ExpectedCount);
   int CountLen(DATAARRAY Data);
   bool FreeBlock();
   bool RoomForAppend(int Len, int Amount);
   int BlockCapacity(PLOGBLOCK Block)  { return Block->IsContinuation() ? BLOCKSIZE-sizeof(BLOCKDESCRIPTOR) : BLOCKSIZE-sizeof(BLOCKDESCRIPTOR)-sizeof(UINT8)-sizeof(LOGTIMESTAMP);   };
   bool ReadSeek(UINT32 Pos, PVOID Data, UINT32 Len);
   bool WriteSeek(UINT32 Pos, PVOID Data, UINT32 Len);
   bool WriteSeekLow(UINT32 Pos, PVOID Data, UINT32 Len);
   bool CopySeek(UINT32 CopyFrom, UINT32 Len, UINT32 CopyTo, bool Reverse);
   bool EraseSeek(UINT32 Pos, UINT32 Len);
   void FlushCache();
private:
   HANDLE hFile;
   CRITICAL_SECTION Lock;
   int MaxBlocks;
   int FirstBlockIndex;
   int LastBlockIndex;
   int BlockCount;
   int LineCount;
   UINT32 LogSize;
   UINT32 ChangeCount;
   UINT32 DeletedLines;
   UINT32 ResetCount;
   DATAARRAY DataBlock;
   int DataBlockLen;
   int DataBlockLineCount;
   int DataBlockIndex;
   bool DataBlockDirty;
   DATAARRAY CompressedBlock;
   int CompressedBlockLen;
   UINT16 BLOCKSIZE;
   UINT16 MAXDATASIZE;
   UINT16 MAXDATAWRITE;
   UINT16 ExtraSize;
   UINT8* ExtraData;
   PLOGBLOCK Block;
   UINT8* RawBlock;
   UINT32 BlockPos;
   UINT32 DataPos;
   UINT32 Flags;
   INT64 EarliestTime;
   UINT8* WrCache;
   UINT32 WrCachePos;
   UINT32 WrCacheLen;
   UINT8* RdCache;
   UINT32 RdCachePos;
   UINT32 RdCacheLen;
   bool StateSaved;
   UINT32 UseCount;
   bool IsClosing;
};

class LOGCURSORIMP
{
public:
   LOGCURSORIMP(LOGIMP& log);
   ~LOGCURSORIMP();
   bool GotoTop();
   bool GotoBottom();
   bool MoveTo(int AbsLine);
   bool MoveBy(int RelLine);
   bool Search(INT64 Time);
   bool Read(INT64& Time, PCHAR& Data, int& Len);
   bool IsNote();
   bool ReadNote(INT64& Time, TCHAR* Note);
   bool IsRecord();
   bool ReadRecord(INT64& Time, TCHAR* Id, TCHAR* Record);
   INT64 ReadTime();
   bool ReadSensor(INT64& Time, TCHAR* Name, TCHAR* Units, UINT8& Quality, double& Num);
   bool ReadSensor(INT64& Time, TCHAR* Name, TCHAR* Units, UINT8& Quality, TCHAR* Num);
   bool AtTop();
   bool AtBottom();
   int GetCurrentLine();
   void SyncPublic();
   bool BeingClosed();
private:
   void Sync();
   void InSync();
   bool Contained();
   LOGIMP& Log;
   int FBI;            // Last known value of Log.FirstBlockIndex
   int BlockIndex;
   int CurrentLine;
   int StartLine;
   int StopLine;
   int Count;
   int StartIndex, StopIndex;
   DATAARRAY DataBlock;
   int DataLen;
   LINEARRAY Lines;
   UINT32 ChangeCount;
   UINT32 DeletedLines;
   UINT32 ResetCount;
   INT64 LastTime;
   char* LastData;
   int LastDataLen;
   bool DataBlockValid;
   int LineCount;
   INT64 BlockTime;
};


#endif
